#define _POSIX_C_SOURCE 200809L
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <errno.h>

typedef  struct word_list  word_list;
struct word_list {
    struct word_list  *next;
    char               text[];
};

void free_words(word_list *next)
{
    while (next) {
        word_list *const curr = next;
        next = next->next;
        free(curr);
    }
}

size_t count_words(word_list *list)
{
    size_t count = 0;
    while (list) {
        list = list->next;
        count++;
    }
    return count;
}

word_list *read_words(FILE *in)
{
    word_list *list = NULL, *curr;
    char      *line = NULL, *word;
    size_t     size = 0;
    ssize_t    bytes;
    size_t     len;

    if (!in) {
        errno = EINVAL;
        return NULL;
    }

    while (1) {

        bytes = getline(&line, &size, in);
        if (bytes == -1)
            break;

        word = line + strspn(line, "\t\n\v\f\r ");
        len = strcspn(line, "\t\n\v\f\r ");
        if (len < 1)
            continue;

        curr = malloc(sizeof (word_list) + len + 1);
        if (!curr) {
            free_words(list);
            errno = ENOMEM;
            return NULL;
        }

        memcpy(curr->text, word, len);
        curr->text[len] = '\0';
        curr->next = list;
        list = curr;
    }

    if (ferror(in) || !feof(in)) {
        free_words(list);
        errno = EIO;
        return NULL;
    }

    return list;
}


static word_list *known = NULL;

static void combo_popup(GtkComboBox *widget, gpointer unused)
{
    GtkWidget    *entry = gtk_bin_get_child(GTK_BIN(widget));
    const char   *prefix = gtk_entry_get_text(GTK_ENTRY(entry));
    size_t        prefix_len = strlen(prefix);
    GtkTreeIter   iterator;
    GtkListStore *combobox_list = GTK_LIST_STORE(gtk_combo_box_get_model(widget));
    word_list    *item;

    gtk_list_store_clear(combobox_list);
    gtk_tree_model_get_iter_first(GTK_TREE_MODEL(combobox_list), &iterator);

    if (prefix_len > 0) {
        /* Rebuild the list of known prefixes. */
        for (item = known; item != NULL; item = item->next)
            if (!strncmp(prefix, item->text, prefix_len)) {
                gtk_list_store_append(combobox_list, &iterator);
                gtk_list_store_set(combobox_list, &iterator, 0, item->text, -1);
            }
    } else {
        /* List all. */
        for (item = known; item != NULL; item = item->next) {
            gtk_list_store_append(combobox_list, &iterator);
            gtk_list_store_set(combobox_list, &iterator, 0, item->text, -1);
        }
    }
}


static void activate(GtkApplication *app, gpointer user_data)
{
    GtkWidget *window;
    GtkWidget *combobox;
    GtkListStore *combobox_list;

    window = gtk_application_window_new(app);
    gtk_window_set_title(GTK_WINDOW(window), "Combo Box Text example");
    gtk_window_set_default_size(GTK_WINDOW(window), 600, 600);
    combobox_list = gtk_list_store_new(1, G_TYPE_STRING);

    combobox = gtk_combo_box_new_with_model_and_entry(GTK_TREE_MODEL(combobox_list));
    gtk_combo_box_set_id_column(GTK_COMBO_BOX(combobox), 0);
    gtk_combo_box_set_entry_text_column(GTK_COMBO_BOX(combobox), 0);
    gtk_combo_box_set_button_sensitivity(GTK_COMBO_BOX(combobox), GTK_SENSITIVITY_ON);
    g_signal_connect(combobox, "popup", G_CALLBACK(combo_popup), window);

    gtk_container_add(GTK_CONTAINER(window), combobox);

    gtk_widget_show_all(window);
}

int main(int argc, char *argv[])
{
    GtkApplication *app;
    int             status;
    size_t          n;

    /* First, we read words from standard input,
       so we have data to match against.
       These particular details are irrelevant. */
    known = read_words(stdin);
    n = count_words(known);
    if (n < 1) {
        fprintf(stderr, "No words in standard input.\n");
        return EXIT_FAILURE;
    }
    fprintf(stderr, "Read %zu words from standard input.\n", n);

    /* Next, we instantiate this example application via activate(). */
    app = gtk_application_new("org.nominal-animal.example", G_APPLICATION_FLAGS_NONE);
    g_signal_connect(app, "activate", G_CALLBACK(activate), NULL);
    status = g_application_run(G_APPLICATION(app), argc, argv);
    g_object_unref(app);

    return status;
}
